@@ -2,7 +2,7 @@
 # -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
 #   Catfish - a versatile file searching tool
 #   Copyright (C) 2007-2012 Christian Dywan <christian@twotoasts.de>
-#   Copyright (C) 2012-2019 Sean Davis <smd.seandavis@gmail.com>
+#   Copyright (C) 2012-2019 Sean Davis <bluesabre@xfce.org>
 #
 #   This program is free software: you can redistribute it and/or modify it
 #   under the terms of the GNU General Public License version 2, as published
@@ -30,6 +30,10 @@ from mimetypes import guess_type
 
 from sys import version_info
 
+import gi
+gi.require_version('GLib', '2.0')  # noqa
+from gi.repository import GLib
+
 try:
     from zeitgeist.client import ZeitgeistDBusInterface
     from zeitgeist.datamodel import Event, TimeRange
@@ -51,6 +55,7 @@ else:
     locate_support = False
 FNULL.close()
 
+
 def get_keyword_list(keywords):
     keywords = keywords.replace(",", " ").strip().lower()
     kwords = []
@@ -63,6 +68,7 @@ def get_keyword_list(keywords):
         kwords.append(keyword.replace("\0", " "))
     return kwords
 
+
 def string_regex(keywords, path):
     """Returns a string with the regular expression containing all combinations
     of the keywords."""
@@ -92,7 +98,8 @@ class CatfishSearchEngine:
     """CatfishSearchEngine is the collection of search backends that are used
     to perform a query.  Each backend is a CatfishSearchMethod"""
 
-    def __init__(self, methods=['zeitgeist', 'locate', 'walk']):
+    def __init__(self, methods=['zeitgeist', 'locate', 'walk'],
+                 exclude_paths=[]):
         """Initialize the CatfishSearchEngine.  Provide a list of methods to
         be included in the search backends.  Available backends include:
 
@@ -118,6 +125,7 @@ class CatfishSearchEngine:
             self.add_method(CatfishSearchMethod_Fulltext)
         if 'walk' in methods:
             self.add_method(CatfishSearchMethod_Walk)
+        self.exclude_paths = exclude_paths
         initialized = []
         for method in self.methods:
             initialized.append(method.method_name)
@@ -165,11 +173,7 @@ class CatfishSearchEngine:
 
         # Path exclusions for efficiency
         exclude = []
-        maybe_exclude = [
-            os.path.expanduser("~/.cache"),
-            os.path.expanduser("~/.gvfs"),
-            "/dev"
-        ]
+        maybe_exclude = self.exclude_paths
         for maybe_path in maybe_exclude:
             if not path.startswith(maybe_path):
                 exclude.append(maybe_path)
@@ -182,11 +186,12 @@ class CatfishSearchEngine:
             logger.debug(
                 "[%i] Starting search method: %s",
                 self.engine_id, method.method_name)
-            for filename in method.run(keywords, path, regex):
+            for filename in method.run(keywords, path, regex,
+                                       self.exclude_paths):
                 if isinstance(filename, str) and path in filename:
                     found_bad = False
                     for filepath in exclude:
-                        if filepath in filename:
+                        if filename.startswith(filepath):
                             if self.stop_time > 0:
                                 logger.debug("Engine is stopped")
                                 return
@@ -268,7 +273,7 @@ class CatfishSearchMethod:
         """Base CatfishSearchMethod Initializer."""
         self.method_name = method_name
 
-    def run(self, keywords, path, regex=False):
+    def run(self, keywords, path, regex=False, exclude_paths=[]):
         """Base CatfishSearchMethod run method."""
         return NotImplemented
 
@@ -291,18 +296,58 @@ class CatfishSearchMethod_Walk(CatfishSearchMethod):
         """Initialize the 'walk' Search Method."""
         CatfishSearchMethod.__init__(self, "walk")
 
-    def run(self, keywords, path, regex=False):
+    def get_dir_list(self, root, dirs, xdg_list, exclude_list):
+        dirs = sorted(dirs, key=lambda s: s.lower())
+
+        # Prioritize: XDG, Visible (Linked), Dotfile (Linked)
+        xdgdirs = []
+        dotdirs = []
+        dotlinks = []
+        notdotdirs = []
+        notdotlinks = []
+        for path in dirs:
+            path = os.path.join(root, path)
+            if not path.endswith("/"):
+                path = path + "/"
+            if path in exclude_list:
+                continue
+            if path in xdg_list:
+                xdgdirs.append(path)
+                continue
+            islink = os.path.islink(path)
+            if os.path.basename(path).startswith("."):
+                if islink:
+                    dotlinks.append(path)
+                else:
+                    dotdirs.append(path)
+            else:
+                if islink:
+                    notdotlinks.append(path)
+                else:
+                    notdotdirs.append(path)
+
+        dirlist = xdgdirs + notdotdirs + notdotlinks + dotdirs + dotlinks
+        return dirlist
+
+    def get_root_list(self, path, xdg_list, exclude_list):
+        # Sort the roots alphabetically
+        roots = [d for d in os.listdir(path)
+                 if os.path.isdir(os.path.join(path, d)) and
+                 os.path.join(path, d) not in exclude_list]
+        roots = sorted(roots, key=lambda s: s.lower())
+        results = []
+        for dirpath in self.get_dir_list(path, roots, xdg_list, exclude_list):
+            results.append(os.path.join(path, dirpath))
+        return results
+
+    def run(self, keywords, path, regex=False, exclude_paths=[]):
         """Run the search method using keywords and path.  regex is not used
         by this search method.
 
         This function is a generator and will yield files as they are found or
         True if still running."""
         exclude = []
-        maybe_exclude = [
-            os.path.expanduser("~/.cache"),
-            os.path.expanduser("~/.gvfs"),
-            "/dev"
-        ]
+        maybe_exclude = exclude_paths
         for maybe_path in maybe_exclude:
             if not path.startswith(maybe_path):
                 exclude.append(maybe_path)
@@ -310,15 +355,48 @@ class CatfishSearchMethod_Walk(CatfishSearchMethod):
         self.running = True
         if isinstance(keywords, str):
             keywords = keywords.replace(',', ' ').strip().split()
-        for root, dirs, files in os.walk(path, False):
-            dirs[:] = [d for d in dirs if os.path.join(root, d) not in exclude]
-            if not self.running:
-                break
-            paths = dirs + files
-            paths.sort()
-            for path in paths:
-                if any(keyword in path.lower() for keyword in keywords):
-                    yield os.path.join(root, path)
+
+        # Enable symbolic link directories, but process once
+        processed_links = []
+
+        # Grab the special directory list to get them precedence
+        xdgdirlist = [
+            GLib.get_user_special_dir(GLib.USER_DIRECTORY_DESKTOP),
+            GLib.get_user_special_dir(GLib.USER_DIRECTORY_DOCUMENTS),
+            GLib.get_user_special_dir(GLib.USER_DIRECTORY_DOWNLOAD),
+            GLib.get_user_special_dir(GLib.USER_DIRECTORY_MUSIC),
+            GLib.get_user_special_dir(GLib.USER_DIRECTORY_PICTURES),
+            GLib.get_user_special_dir(
+                GLib.USER_DIRECTORY_PUBLIC_SHARE),
+            GLib.get_user_special_dir(GLib.USER_DIRECTORY_TEMPLATES),
+            GLib.get_user_special_dir(GLib.USER_DIRECTORY_VIDEOS),
+        ]
+
+        for path in self.get_root_list(path, xdgdirlist, exclude):
+            for root, dirs, files in os.walk(top=path, topdown=True,
+                                             onerror=None,
+                                             followlinks=True):
+                # Bail once the search has been canceled
+                if not self.running:
+                    break
+
+                # Check if we've already processed symbolic paths
+                if os.path.islink(root):
+                    realpath = os.path.realpath(root)
+                    if realpath in processed_links:
+                        yield True
+                        continue
+                    processed_links.append(realpath)
+
+                # Prioritize and drop excluded paths
+                dirs[:] = self.get_dir_list(root, dirs, xdgdirlist, exclude)
+
+                paths = dirs + files
+                paths.sort()
+                for path in paths:
+                    if any(keyword in path.lower() for keyword in keywords):
+                        yield os.path.join(root, path)
+                yield True
             yield True
         yield False
 
@@ -344,7 +422,7 @@ class CatfishSearchMethod_Fulltext(CatfishSearchMethod):
         self.running = False
         self.exact = False
 
-    def run(self, keywords, path, regex=False): # noqa
+    def run(self, keywords, path, regex=False, exclude_paths=[]):  # noqa
         """Run the search method using keywords and path.  regex is not used
         by this search method.
 
@@ -432,7 +510,7 @@ class CatfishSearchMethod_Zeitgeist(CatfishSearchMethod):
         """Initialize the Zeitgeist SearchMethod."""
         CatfishSearchMethod.__init__(self, "zeitgeist")
 
-    def run(self, keywords, path, regex=False):
+    def run(self, keywords, path, regex=False, exclude_paths=[]):
         """Run the Zeitgeist SearchMethod."""
         self.stop_search = False
         event_template = Event()
@@ -488,9 +566,9 @@ class CatfishSearchMethodExternal(CatfishSearchMethod):
 
     def assemble_query(self, keywords, path):
         """Base assemble_query method."""
-        return None
+        return False
 
-    def run(self, keywords, path, regex=False):
+    def run(self, keywords, path, regex=False, exclude_paths=[]):
         """Run the search method using keywords and path.
 
         This function returns the process.stdout generator and will yield files
@@ -512,7 +590,8 @@ class CatfishSearchMethodExternal(CatfishSearchMethod):
     def process_output(self, output):
         """Return the output text."""
         if isinstance(output, io.BufferedReader):
-            return map(lambda s: s.decode(encoding='UTF8', errors='replace').strip(),
+            return map(lambda s: s.decode(encoding='UTF8',
+                                          errors='replace').strip(),
                        output.readlines())
         else:
             return output
@@ -547,8 +626,33 @@ class CatfishSearchMethod_Locate(CatfishSearchMethodExternal):
     def __init__(self):
         """Initialize the Locate SearchMethod."""
         CatfishSearchMethodExternal.__init__(self, "locate")
-        self.command = ["locate", "-i", "%path*%keywords*"]
+        self.caps = self.get_capabilities()
+        if self.caps["existing"]:
+            self.command = ["locate", "-i", "%path*%keywords*", "--existing"]
+        else:
+            self.command = ["locate", "-i", "%path*%keywords*"]
+
+    def get_capabilities(self):
+        caps = {
+            "existing": False,
+            "regex": False
+        }
+        try:
+            details = subprocess.check_output(["locate", "--help"])
+            details = details.decode("utf-8")
+            if "--existing" in details:
+                caps["existing"] = True
+            if "--regex" in details or "--regexp" in details:
+                caps["regex"] = True
+
+        except subprocess.CalledProcessError:
+            pass
+        return caps
 
     def assemble_query(self, keywords, path):
         """Assemble the search query."""
-        return ["locate", "-i", "%path*", str(keywords)]
+        if self.caps["regex"]:
+            return ["locate", "--regex", "--basename", "-i",
+                    "{}".format(string_regex(keywords, path))]
+        else:
+            return ["locate", "-i", "%path*", str(keywords)]
